import { makeLegalIdentifier, createFilter } from '@rollup/pluginutils';
import path from 'path';
import Concat from 'concat-with-sourcemaps';
import cssnano from 'cssnano';
import PQueue from 'p-queue';
import postcss from 'postcss';
import postcssrc from 'postcss-load-config';
import fs from 'fs-extra';
import { SourceMapConsumer } from 'source-map';
import resolveAsync$1 from 'resolve';
import { createHash } from 'crypto';
import valueParser from 'postcss-value-parser';
import { lookup } from 'mime-types';
import modulesValues from 'postcss-modules-values';
import localByDefault from 'postcss-modules-local-by-default';
import extractImports from 'postcss-modules-extract-imports';
import modulesScope from 'postcss-modules-scope';
import { extractICSS, replaceSymbols, replaceValueSymbols } from 'icss-utils';

const isAbsolutePath = path => /^(?:\/|(?:[A-Za-z]:)?[/\\|])/.test(path);
const isRelativePath = path => /^\.?\.[/\\]/.test(path);
function normalizePath(...paths) {
  const f = path.join(...paths).replace(/\\/g, "/");
  if (/^\.[/\\]/.test(paths[0])) return `./${f}`;
  return f;
}
const resolvePath = (...paths) => normalizePath(path.resolve(...paths));
const relativePath = (from, to) => normalizePath(path.relative(from, to));
const humanlizePath = file => relativePath(process.cwd(), file);

function _classPrivateFieldGet(receiver, privateMap) {
  var descriptor = privateMap.get(receiver);

  if (!descriptor) {
    throw new TypeError("attempted to get private field on non-instance");
  }

  if (descriptor.get) {
    return descriptor.get.call(receiver);
  }

  return descriptor.value;
}

function _classPrivateFieldSet(receiver, privateMap, value) {
  var descriptor = privateMap.get(receiver);

  if (!descriptor) {
    throw new TypeError("attempted to set private field on non-instance");
  }

  if (descriptor.set) {
    descriptor.set.call(receiver, value);
  } else {
    if (!descriptor.writable) {
      throw new TypeError("attempted to set read only private field");
    }

    descriptor.value = value;
  }

  return value;
}

const hashRe = /\[hash(?::(\d+))?]/;
const firstExtRe = /(?<!^|[/\\])(\.[^\s.]+)/i;
const dataURIRe = /data:[^\n\r;]+?(?:;charset=[^\n\r;]+?)?;base64,([\d+/A-Za-z]+={0,2})/;

const isNullish = v => {
  const _v = v !== null && v !== void 0 ? v : undefined;

  if (typeof _v === "undefined") return true;
  return false;
};
const nullishFilter = v => !isNullish(v);
const denullifyObject = o => Object.entries(o).filter(([, v]) => nullishFilter(v)).reduce((acc, [k, v]) => ({ ...acc,
  [k]: v
}), {});
const booleanFilter = v => Boolean(v);

const mapBlockRe = /(?:\n|\r\n)?\/\*[#*@]+?\s*?sourceMappingURL\s*?=\s*?(\S+)\s*?\*+?\//gm;
const mapLineRe = /(?:\n|\r\n)?\/\/[#@]+?\s*?sourceMappingURL\s*?=\s*?(\S+)\s*?$/gm;
async function getMap(code, id) {
  var _ref, _mapBlockRe$exec, _dataURIRe$exec;

  const [, data] = (_ref = (_mapBlockRe$exec = mapBlockRe.exec(code)) !== null && _mapBlockRe$exec !== void 0 ? _mapBlockRe$exec : mapLineRe.exec(code)) !== null && _ref !== void 0 ? _ref : [];
  if (!data) return;
  const [, uriMap] = (_dataURIRe$exec = dataURIRe.exec(data)) !== null && _dataURIRe$exec !== void 0 ? _dataURIRe$exec : [];
  if (uriMap) return Buffer.from(uriMap, "base64").toString();
  if (!id) throw new Error("Extracted map detected, but no ID is provided");
  const mapFileName = path.resolve(path.dirname(id), data);

  try {
    return await fs.readFile(mapFileName, "utf8");
  } catch {
    return;
  }
}
const stripMap = code => code.replace(mapBlockRe, "").replace(mapLineRe, "");

class MapModifier {
  constructor(map) {
    _map.set(this, {
      writable: true,
      value: void 0
    });

    if (typeof map === "string") try {
      _classPrivateFieldSet(this, _map, JSON.parse(map));
    } catch {
      /* noop */
    } else _classPrivateFieldSet(this, _map, map && { ...map
    });
  }

  modify(f) {
    if (isNullish(_classPrivateFieldGet(this, _map))) return this;
    f(_classPrivateFieldGet(this, _map));
    return this;
  }

  modifySources(op) {
    if (isNullish(_classPrivateFieldGet(this, _map))) return this;
    if (_classPrivateFieldGet(this, _map).sources) _classPrivateFieldGet(this, _map).sources = _classPrivateFieldGet(this, _map).sources.map(op);
    return this;
  }

  resolve(dir) {
    return this.modifySources(source => {
      if (source === "<no source>") return source;
      return resolvePath(dir, source);
    });
  }

  relative(dir = process.cwd()) {
    return this.modifySources(source => {
      if (source === "<no source>") return source;
      if (isAbsolutePath(source)) return relativePath(dir, source);else return normalizePath(source);
    });
  }

  toObject() {
    return _classPrivateFieldGet(this, _map);
  }

  toString() {
    if (isNullish(_classPrivateFieldGet(this, _map))) return _classPrivateFieldGet(this, _map);
    return JSON.stringify(_classPrivateFieldGet(this, _map));
  }

  async toConsumer() {
    if (isNullish(_classPrivateFieldGet(this, _map))) return _classPrivateFieldGet(this, _map);
    return new SourceMapConsumer(_classPrivateFieldGet(this, _map));
  }

  toCommentData() {
    const map = this.toString();
    if (isNullish(map)) return "";
    const sourceMapData = Buffer.from(map).toString("base64");
    return `\n/*# sourceMappingURL=data:application/json;base64,${sourceMapData} */`;
  }

  toCommentFile(fileName) {
    if (isNullish(_classPrivateFieldGet(this, _map))) return "";
    return `\n/*# sourceMappingURL=${fileName} */`;
  }

}

var _map = new WeakMap();

const mm = map => new MapModifier(map);

var resolveAsync = (async (id, options = {}) => new Promise((resolve, reject) => {
  resolveAsync$1(id, options, (err, res) => err ? reject(err) : resolve(res));
}));

var hasher = (data => createHash("sha256").update(data).digest("hex"));

var safeId = ((id, ...salt) => {
  const hash = hasher([id, ...salt].join(":")).slice(0, 8);
  return makeLegalIdentifier(`${id}_${hash}`);
});

const resolve = async (url, basedir, extensions) => {
  const options = {
    basedir,
    extensions
  };
  let from;

  try {
    from = await resolveAsync(url, options);
  } catch {
    from = await resolveAsync(`./${url}`, options);
  }

  return {
    from,
    source: await fs.readFile(from)
  };
};

const name = "styles-import";
const plugin = postcss.plugin(name, (options = {}) => async (css, res) => {
  var _css$source, _options$resolve, _options$alias, _options$extensions;

  if (!((_css$source = css.source) === null || _css$source === void 0 ? void 0 : _css$source.input.file)) return;
  const resolve$1 = (_options$resolve = options.resolve) !== null && _options$resolve !== void 0 ? _options$resolve : resolve;
  const alias = (_options$alias = options.alias) !== null && _options$alias !== void 0 ? _options$alias : {};
  const extensions = (_options$extensions = options.extensions) !== null && _options$extensions !== void 0 ? _options$extensions : [".css", ".pcss", ".postcss", ".sss"];
  const opts = res.opts && { ...res.opts
  };
  opts === null || opts === void 0 ? true : delete opts.map;
  const {
    file
  } = css.source.input;
  const importList = [];
  const basedir = path.dirname(file);
  css.walkAtRules(/^import$/i, importRule => {
    // Top level only
    if (importRule.parent.type !== "root") {
      importRule.warn(res, "`@import` should be top level");
      return;
    } // Child nodes should not exist


    if (importRule.nodes) {
      importRule.warn(res, "`@import` was not terminated correctly");
      return;
    }

    const [urlNode] = valueParser(importRule.params).nodes; // No URL detected

    if (!urlNode || urlNode.type !== "string" && urlNode.type !== "function") {
      importRule.warn(res, `No URL in \`${importRule.toString()}\``);
      return;
    }

    let url = "";

    if (urlNode.type === "string") {
      url = urlNode.value;
    } else if (urlNode.type === "function") {
      var _urlNode$nodes$;

      // Invalid function
      if (!/^url$/i.test(urlNode.value)) {
        importRule.warn(res, `Invalid \`url\` function in \`${importRule.toString()}\``);
        return;
      }

      const isString = ((_urlNode$nodes$ = urlNode.nodes[0]) === null || _urlNode$nodes$ === void 0 ? void 0 : _urlNode$nodes$.type) === "string";
      url = isString ? urlNode.nodes[0].value : valueParser.stringify(urlNode.nodes);
    }

    url = url.replace(/^\s+|\s+$/g, ""); // Resolve aliases

    for (const [from, to] of Object.entries(alias)) {
      if (!url.startsWith(from)) continue;
      url = normalizePath(to) + url.slice(from.length);
    } // Empty url


    if (url.length === 0) {
      importRule.warn(res, `Empty URL in \`${importRule.toString()}\``);
      return;
    }

    importList.push({
      importRule,
      url
    });
  });

  for await (const {
    importRule,
    url
  } of importList) {
    try {
      const {
        source,
        from
      } = await resolve$1(url, basedir, extensions);

      if (!(source instanceof Uint8Array) || typeof from !== "string") {
        importRule.warn(res, `Incorrectly resolved \`@import\` in \`${importRule.toString()}\``);
        continue;
      }

      if (normalizePath(from) === normalizePath(file)) {
        importRule.warn(res, `\`@import\` loop in \`${importRule.toString()}\``);
        continue;
      }

      res.messages.push({
        plugin: name,
        type: "dependency",
        file: from
      });
      const imported = await postcss(plugin(options)).process(source, { ...opts,
        from
      });
      res.messages.unshift(...imported.messages);
      if (!imported.root) importRule.remove();else importRule.replaceWith(imported.root);
    } catch {
      importRule.warn(res, `Unresolved \`@import\` in \`${importRule.toString()}\``);
    }
  }
});

const resolve$1 = async (url, basedir) => {
  const options = {
    basedir
  };
  let from;

  try {
    from = await resolveAsync(url, options);
  } catch {
    from = await resolveAsync(`./${url}`, options);
  }

  return {
    from,
    source: await fs.readFile(from)
  };
};

var generateName = ((placeholder, file, source) => {
  const hash = hasher(`${path.basename(file)}:${Buffer.from(source).toString()}`);
  const match = hashRe.exec(placeholder);
  const hashLen = match && Number.parseInt(match[1]);
  return placeholder.replace("[dir]", path.basename(path.dirname(file))).replace("[name]", path.basename(file, path.extname(file))).replace("[extname]", path.extname(file)).replace(".[ext]", path.extname(file)).replace("[ext]", path.extname(file).slice(1)).replace(hashRe, hashLen ? hash.slice(0, hashLen) : hash.slice(0, 8));
});

const urlFuncRe = /^url$/i;
const imageSetFuncRe = /^(?:-webkit-)?image-set$/i;
const isDeclWithUrl = decl => /^(?:url|(?:-webkit-)?image-set)\(/i.test(decl.value);
const walkUrls = (parsed, callback) => {
  parsed.walk(node => {
    if (node.type !== "function") return;

    if (urlFuncRe.test(node.value)) {
      const {
        nodes
      } = node;
      const [urlNode] = nodes;
      const url = (urlNode === null || urlNode === void 0 ? void 0 : urlNode.type) === "string" ? urlNode.value : valueParser.stringify(nodes);
      callback(url.replace(/^\s+|\s+$/g, ""), urlNode);
      return;
    }

    if (imageSetFuncRe.test(node.value)) {
      for (const nNode of node.nodes) {
        if (nNode.type === "string") {
          callback(nNode.value.replace(/^\s+|\s+$/g, ""), nNode);
          continue;
        }

        if (nNode.type === "function" && urlFuncRe.test(nNode.value)) {
          const {
            nodes
          } = nNode;
          const [urlNode] = nodes;
          const url = (urlNode === null || urlNode === void 0 ? void 0 : urlNode.type) === "string" ? urlNode.value : valueParser.stringify(nodes);
          callback(url.replace(/^\s+|\s+$/g, ""), urlNode);
          continue;
        }
      }
    }
  });
};

var inlineFile = ((file, source) => {
  const mime = lookup(file) || "application/octet-stream";
  const data = Buffer.from(source).toString("base64");
  return `data:${mime};base64,${data}`;
});

const name$1 = "styles-url";
const plugin$1 = postcss.plugin(name$1, (options = {}) => async (css, res) => {
  var _css$source, _options$inline, _options$publicPath, _options$assetDir, _options$resolve, _options$alias, _options$hash, _css$source$input$map;

  if (!((_css$source = css.source) === null || _css$source === void 0 ? void 0 : _css$source.input.file)) return;
  const inline = (_options$inline = options.inline) !== null && _options$inline !== void 0 ? _options$inline : false;
  const publicPath = (_options$publicPath = options.publicPath) !== null && _options$publicPath !== void 0 ? _options$publicPath : "./";
  const assetDir = (_options$assetDir = options.assetDir) !== null && _options$assetDir !== void 0 ? _options$assetDir : ".";
  const resolve = (_options$resolve = options.resolve) !== null && _options$resolve !== void 0 ? _options$resolve : resolve$1;
  const alias = (_options$alias = options.alias) !== null && _options$alias !== void 0 ? _options$alias : {};
  const placeholder = ((_options$hash = options.hash) !== null && _options$hash !== void 0 ? _options$hash : true) ? typeof options.hash === "string" ? options.hash : "assets/[name]-[hash][extname]" : "assets/[name][extname]";
  const {
    file
  } = css.source.input;
  const map = await mm((_css$source$input$map = css.source.input.map) === null || _css$source$input$map === void 0 ? void 0 : _css$source$input$map.text).resolve(path.dirname(file)).toConsumer();
  const urlList = [];
  const imported = res.messages.filter(msg => msg.type === "dependency").map(msg => msg.file);
  css.walkDecls(decl => {
    if (!isDeclWithUrl(decl)) return;
    const parsed = valueParser(decl.value);
    walkUrls(parsed, (url, node) => {
      var _decl$source, _decl$source2;

      // Resolve aliases
      for (const [from, to] of Object.entries(alias)) {
        if (!url.startsWith(from)) continue;
        url = normalizePath(to) + url.slice(from.length);
      } // Empty URL


      if (!node || url.length === 0) {
        decl.warn(res, `Empty URL in \`${decl.toString()}\``);
        return;
      } // Skip Data URI


      if (dataURIRe.test(url)) return; // Skip Web URLs

      if (!isAbsolutePath(url)) {
        try {
          new URL(url);
          return;
        } catch {// Is not a Web URL, continuing
        }
      } // Use PostCSS imports


      if (((_decl$source = decl.source) === null || _decl$source === void 0 ? void 0 : _decl$source.input.file) && imported.includes(decl.source.input.file)) {
        const basedir = path.dirname(decl.source.input.file);
        urlList.push({
          node,
          url,
          decl,
          parsed,
          basedir
        });
        return;
      } // Use SourceMap


      if ((_decl$source2 = decl.source) === null || _decl$source2 === void 0 ? void 0 : _decl$source2.start) {
        const pos = decl.source.start;
        const realPos = map === null || map === void 0 ? void 0 : map.originalPositionFor(pos);
        const basedir = (realPos === null || realPos === void 0 ? void 0 : realPos.source) && path.dirname(realPos.source);

        if (basedir) {
          urlList.push({
            node,
            url,
            decl,
            parsed,
            basedir
          });
          return;
        }
      } // Use current file


      const basedir = path.dirname(file);
      urlList.push({
        node,
        url,
        decl,
        parsed,
        basedir
      });
    });
  });
  map === null || map === void 0 ? void 0 : map.destroy();
  const usedNames = new Map();

  for await (const {
    node,
    url,
    decl,
    parsed,
    basedir
  } of urlList) {
    try {
      const {
        source,
        from
      } = await resolve(url, basedir);

      if (!(source instanceof Uint8Array) || typeof from !== "string") {
        decl.warn(res, `Incorrectly resolved URL \`${url}\` in \`${decl.toString()}\``);
        continue;
      }

      res.messages.push({
        plugin: name$1,
        type: "dependency",
        file: from
      });

      if (inline) {
        node.type = "string";
        node.value = inlineFile(from, source);
      } else {
        let safeTo, to;
        to = safeTo = normalizePath(generateName(placeholder, from, source)); // Avoid file overrides

        for (let i = 1; usedNames.has(safeTo) && usedNames.get(safeTo) !== from; i++) safeTo = firstExtRe.test(to) ? to.replace(firstExtRe, `${i}$1`) : `${to}${i}`;

        to = safeTo;
        usedNames.set(to, from);
        node.type = "string";
        node.value = publicPath + (/[/\\]$/.test(publicPath) ? "" : "/") + path.basename(to);
        to = normalizePath(assetDir, safeTo);
        res.messages.push({
          plugin: name$1,
          type: "asset",
          to,
          source
        });
      }

      decl.value = parsed.toString();
    } catch {
      decl.warn(res, `Unresolved URL \`${url}\` in \`${decl.toString()}\``);
    }
  }
});

var generateScopedNameDefault = ((placeholder = "[name]_[local]__[hash:8]") => (name, file, css) => {
  const hash = hasher(`${path.basename(file)}:${css}`);
  const match = hashRe.exec(placeholder);
  const hashLen = match && Number.parseInt(match[1]);
  return makeLegalIdentifier(placeholder.replace("[dir]", path.basename(path.dirname(file))).replace("[name]", path.basename(file, path.extname(file))).replace("[local]", name).replace(hashRe, hashLen ? hash.slice(0, hashLen) : hash));
});

var postcssModules = (options => {
  const opts = {
    mode: "local",
    ...options,
    generateScopedName: typeof options.generateScopedName === "function" ? options.generateScopedName : generateScopedNameDefault(options.generateScopedName)
  };
  return [modulesValues(), localByDefault({
    mode: opts.mode
  }), extractImports({
    failOnWrongOrder: opts.failOnWrongOrder
  }), modulesScope({
    exportGlobals: opts.exportGlobals,
    generateScopedName: opts.generateScopedName
  })];
});

const load = async (url, file, extensions, processor, opts) => {
  let from;
  const options = {
    basedir: path.dirname(file),
    extensions
  };

  try {
    from = await resolveAsync(url, options);
  } catch {
    from = await resolveAsync(`./${url}`, options);
  }

  const source = await fs.readFile(from);
  const {
    messages
  } = await processor.process(source, { ...opts,
    from
  });
  return messages.filter(msg => msg.type === "icss").reduce((acc, msg) => ({ ...acc,
    ...msg.replacements
  }), {});
};

async function resolve$2 (icssImports, load, file, extensions, processor, opts) {
  return Object.entries(icssImports).reduce(async (acc, [url, values]) => {
    const exports = await load(url, file, extensions, processor, opts);
    const mappedValues = Object.entries(values).reduce((acc, [k, v]) => ({ ...acc,
      [k]: exports[v]
    }), {});
    return { ...(await acc),
      ...mappedValues
    };
  }, Promise.resolve({}));
}

const name$2 = "styles-icss";
const plugin$2 = postcss.plugin(name$2, (options = {}) => async (css, res) => {
  var _css$source, _options$load, _options$extensions, _res$opts;

  if (!((_css$source = css.source) === null || _css$source === void 0 ? void 0 : _css$source.input.file)) return;
  if (!res.processor) return;
  const load$1 = (_options$load = options.load) !== null && _options$load !== void 0 ? _options$load : load;
  const extensions = (_options$extensions = options.extensions) !== null && _options$extensions !== void 0 ? _options$extensions : [".css", ".pcss", ".postcss", ".sss"];
  const opts = res.opts && { ...res.opts
  };
  opts === null || opts === void 0 ? true : delete opts.map;
  const {
    icssExports,
    icssImports
  } = extractICSS(css);
  const imports = await resolve$2(icssImports, load$1, css.source.input.file, extensions, res.processor, opts);
  replaceSymbols(css, imports);
  const replacements = Object.entries(icssExports).reduce((acc, [k, v]) => ({ ...acc,
    [k]: replaceValueSymbols(v, imports)
  }), {});
  res.messages.push({
    plugin: name$2,
    type: "icss",
    replacements
  });
  if (typeof options.getReplacements === "function") options.getReplacements(css.source.input.file, replacements, (_res$opts = res.opts) === null || _res$opts === void 0 ? void 0 : _res$opts.to);
});

const name$3 = "styles-noop";

function noop() {
  /* noop */
}

var postcssNoop = postcss.plugin(name$3, () => noop);

async function loadConfig(id, config) {
  var _config$ctx;

  const configPath = typeof config === "object" && config.path ? path.resolve(config.path) : path.dirname(id);
  const context = {
    file: {
      extname: path.extname(id),
      dirname: path.dirname(id),
      basename: path.basename(id)
    },
    options: typeof config === "object" ? (_config$ctx = config.ctx) !== null && _config$ctx !== void 0 ? _config$ctx : {} : {}
  };
  return postcssrc(context, configPath).catch(error => {
    if (!/no postcss config found/i.test(error.message)) throw error;
    return {};
  });
}

function isAutoModules(am, id) {
  if (typeof am === "function") return am(id);
  if (am instanceof RegExp) return am.test(id);
  return am && /\.module\.[A-Za-z]+$/.test(id);
}

const loader = {
  name: "postcss",
  alwaysProcess: true,

  async process({
    code,
    map,
    extracted
  }) {
    var _options$postcss$plug, _config$plugins, _config$options, _options$to, _res$map, _modulesExports$this$;

    const {
      options
    } = this;
    const config = await loadConfig(this.id, options.config);
    const plugins = [...[options.import && plugin({
      extensions: options.extensions,
      ...options.import
    }), options.url && plugin$1({
      inline: Boolean(options.inject),
      ...options.url
    })].filter(booleanFilter), ...((_options$postcss$plug = options.postcss.plugins) !== null && _options$postcss$plug !== void 0 ? _options$postcss$plug : []), ...((_config$plugins = config.plugins) !== null && _config$plugins !== void 0 ? _config$plugins : [])];
    const autoModules = isAutoModules(options.autoModules, this.id);
    const supportModules = Boolean(options.modules || autoModules);
    const modulesExports = {};
    const postcssOpts = { ...denullifyObject((_config$options = config.options) !== null && _config$options !== void 0 ? _config$options : {}),
      ...denullifyObject(options.postcss),
      from: this.id,
      to: (_options$to = options.to) !== null && _options$to !== void 0 ? _options$to : this.id,
      map: {
        inline: false,
        annotation: false,
        sourcesContent: this.sourceMap ? this.sourceMap.content : true,
        prev: mm(map).relative(path.dirname(this.id)).toObject()
      }
    };
    delete postcssOpts.plugins;

    if (supportModules) {
      const modulesOptions = typeof options.modules === "object" ? options.modules : {};
      plugins.push(...postcssModules({
        generateScopedName:  undefined,
        failOnWrongOrder: true,
        ...modulesOptions
      }), plugin$2({
        extensions: options.extensions,

        getReplacements(file, replacements, out) {
          modulesExports[file] = replacements;

          if (typeof options.modules === "object" && typeof options.modules.getReplacements === "function") {
            return options.modules.getReplacements(file, replacements, out);
          }
        }

      }));
    }

    if (options.minimize) plugins.push(cssnano(typeof options.minimize === "object" ? options.minimize : {})); // Avoid PostCSS warning

    if (plugins.length === 0) plugins.push(postcssNoop);
    const res = await postcss(plugins).process(code, postcssOpts);

    for (const warning of res.warnings()) this.warn({
      name: warning.plugin,
      message: warning.text
    });

    const deps = res.messages.filter(msg => msg.type === "dependency");

    for (const dep of deps) this.deps.add(normalizePath(dep.file));

    const assets = res.messages.filter(msg => msg.type === "asset");

    for (const asset of assets) this.assets.set(asset.to, asset.source);

    map = mm((_res$map = res.map) === null || _res$map === void 0 ? void 0 : _res$map.toJSON()).resolve(path.dirname(postcssOpts.to)).toString();
    if (!options.extract && this.sourceMap) res.css += mm(map).modify(map => void delete map.file).relative().toCommentData();
    if (options.emit) return {
      code: res.css,
      map
    };
    const reservedWords = new Set();

    const saferId = id => safeId(id, humanlizePath(this.id));

    const cssExportName = "css";
    reservedWords.add(cssExportName);
    const cssVarName = saferId(cssExportName);
    const modulesVarName = saferId("modules");
    const output = [`const ${cssVarName} = ${JSON.stringify(res.css)}`, `const ${modulesVarName} = ${JSON.stringify((_modulesExports$this$ = modulesExports[this.id]) !== null && _modulesExports$this$ !== void 0 ? _modulesExports$this$ : {})}`, `export const ${cssExportName} = ${cssVarName}`, `export default ${supportModules ? modulesVarName : cssVarName}`];

    if (options.namedExports) {
      const json = modulesExports[this.id];
      const getClassName = typeof options.namedExports === "function" ? options.namedExports : name => makeLegalIdentifier(reservedWords.has(name) ? `_${name}` : name);

      for (const name in json) {
        const newName = getClassName(name);
        if (name !== newName) this.warn(`Exported \`${name}\` as \`${newName}\` in ${humanlizePath(this.id)}`);
        if (!json[newName]) json[newName] = json[name];
        output.push(`export const ${newName} = ${JSON.stringify(json[name])}`);
      }
    }

    if (options.extract) extracted = {
      id: this.id,
      css: res.css,
      map
    };

    if (options.inject) {
      if (typeof options.inject === "function") {
        output.push(options.inject(cssVarName, this.id));
      } else {
        const injectorVarName = saferId("injector");
        const injectorId = await resolveAsync("./inject-css", {
          basedir: path.join( __dirname, "runtime")
        });
        const injectorData = typeof options.inject === "object" ? `,${JSON.stringify(options.inject)}` : "";
        output.push(`import ${injectorVarName} from '${normalizePath(injectorId)}'`, `${injectorVarName}(${cssVarName}${injectorData})`);
      }
    }

    code = `${output.join(";\n")};`;
    return {
      code,
      map,
      extracted
    };
  }

};

const loader$1 = {
  name: "sourcemap",
  alwaysProcess: true,

  async process({
    code,
    map
  }) {
    var _await$getMap;

    map = (_await$getMap = await getMap(code, this.id)) !== null && _await$getMap !== void 0 ? _await$getMap : map;
    return {
      code: stripMap(code),
      map
    };
  }

};

var loadModule = (async (moduleId, basedir = process.cwd()) => {
  if (typeof moduleId !== "string") return;

  try {
    return require(moduleId);
  } catch {
    /* noop */
  }

  try {
    return require(await resolveAsync(moduleId, {
      basedir
    }));
  } catch {
    /* noop */
  }

  try {
    return require(await resolveAsync(`./${moduleId}`, {
      basedir
    }));
  } catch {
    return;
  }
});

const allSassIDs = ["node-sass", "sass"];
const idFmt = allSassIDs.map((id, i, arr) => {
  const newId = `\`${id}\``;
  if (i === arr.length - 2) return newId;
  if (i === arr.length - 1) return `or ${newId}`;
  return `${newId},`;
}).join(" ");
async function loadSass(impl) {
  // Loading provided implementation
  if (impl) {
    const provided = await loadModule(impl);
    if (provided) return [provided, impl];
    throw new Error(`Could not load \`${impl}\` Sass implementation`);
  } // Loading one of the supported modules


  for await (const id of allSassIDs) {
    const sass = await loadModule(id);
    if (sass) return [sass, id];
  }

  throw new Error(`You need to install ${idFmt} package in order to process Sass files`);
}

const isModule = url => /^~[\d@A-Za-z]\S+/.test(url);
function getUrlOfPartial(url) {
  const parsedUrl = path.parse(url);
  return `${normalizePath(parsedUrl.dir)}/_${parsedUrl.base}`;
}
function normalizeUrl(url) {
  if (isModule(url)) return normalizePath(url.slice(1));
  if (isAbsolutePath(url) || isRelativePath(url)) return normalizePath(url);
  return `./${normalizePath(url)}`;
}

const importer = (url, importer, done) => {
  // Do not add `.css` extension in order to inline the file
  const finishImport = id => done({
    file: id.replace(/\.css$/i, "")
  }); // Pass responsibility back to other custom importers


  const next = () => done(null);

  if (!isModule(url)) return next();
  const moduleUrl = normalizeUrl(url);
  const partialUrl = getUrlOfPartial(moduleUrl);
  const options = {
    basedir: path.dirname(importer),
    extensions: [".scss", ".sass", ".css"]
  }; // Give precedence to importing a partial

  resolveAsync(partialUrl, options).then(finishImport).catch(() => {
    resolveAsync(moduleUrl, options).then(finishImport).catch(next);
  });
};

const loader$2 = {
  name: "sass",
  test: /\.(sass|scss)$/i,

  async process({
    code,
    map
  }) {
    var _options$fibers, _options$data, _options$importer;

    const {
      options
    } = this;
    const [sass, type] = await loadSass(options.impl); // `fibers` doesn't work in testing

    const useFibers = (_options$fibers = options.fibers) !== null && _options$fibers !== void 0 ? _options$fibers : type === "sass" && true;
    const fiber = useFibers ? await loadModule("fibers") : undefined;

    const render = async options => new Promise((resolve, reject) => {
      sass.render(options, (err, css) => err ? reject(err) : resolve(css));
    }); // Remove non-Sass options


    delete options.fibers;
    delete options.impl; // node-sass won't produce source maps if the `data`
    // option is used and `sourceMap` option is not a string.
    //
    // In case it is a string, `sourceMap` option
    // should be a path where the source map is written.
    //
    // But since we're using the `data` option,
    // the source map will not actually be written, but
    // all paths in sourcemap's sources will be relative to that path.

    const res = await render({ ...options,
      file: this.id,
      data: ((_options$data = options.data) !== null && _options$data !== void 0 ? _options$data : "") + code,
      indentedSyntax: /\.sass$/i.test(this.id),
      sourceMap: this.id,
      omitSourceMapUrl: true,
      sourceMapContents: true,
      importer: [importer].concat((_options$importer = options.importer) !== null && _options$importer !== void 0 ? _options$importer : []),
      fiber
    });
    const deps = res.stats.includedFiles;

    for (const dep of deps) this.deps.add(normalizePath(dep));

    return {
      code: Buffer.from(res.css).toString(),
      map: res.map ? Buffer.from(res.map).toString() : map
    };
  }

};

const loader$3 = {
  name: "stylus",
  test: /\.(styl|stylus)$/i,

  async process({
    code,
    map
  }) {
    var _this$options$paths, _style$sourcemap, _mm$toString;

    const stylus = await loadModule("stylus");
    if (!stylus) throw new Error("You need to install `stylus` package in order to process Stylus files");
    const basePath = normalizePath(path.dirname(this.id));
    const style = stylus(code, { ...this.options
    }).set("filename", this.id).set("paths", [`${basePath}/node_modules`, basePath].concat((_this$options$paths = this.options.paths) !== null && _this$options$paths !== void 0 ? _this$options$paths : [])).set("sourcemap", {
      comment: false,
      basePath
    });

    const render = async () => new Promise((resolve, reject) => {
      style.render((err, css) => err ? reject(err) : resolve(css));
    });

    code = await render();
    const deps = style.deps();

    for (const dep of deps) this.deps.add(normalizePath(dep)); // We have to manually modify the `sourcesContent` field
    // since stylus compiler doesn't support it yet


    if (((_style$sourcemap = style.sourcemap) === null || _style$sourcemap === void 0 ? void 0 : _style$sourcemap.sources) && !style.sourcemap.sourcesContent) {
      style.sourcemap.sourcesContent = await Promise.all(style.sourcemap.sources.map(async source => {
        try {
          const file = normalizePath(basePath, source);
          return await fs.readFile(file, "utf8");
        } catch {
          return null;
        }
      }));
    }

    map = (_mm$toString = mm(style.sourcemap).toString()) !== null && _mm$toString !== void 0 ? _mm$toString : map;
    return {
      code,
      map
    };
  }

};

const getStylesFileManager = less => new class extends less.AbstractFileManager {
  supports() {
    return true;
  }

  async loadFile(filename, basedir) {
    const options = {
      basedir,
      extensions: [".less", ".css"]
    };
    const url = normalizeUrl(filename);
    const partialUrl = getUrlOfPartial(url); // Give precedence to importing a partial

    let id;

    try {
      id = await resolveAsync(partialUrl, options);
    } catch {
      id = await resolveAsync(url, options);
    }

    return {
      filename: id,
      contents: await fs.readFile(id, "utf8")
    };
  }

}();

const importer$1 = {
  install(less, pluginManager) {
    pluginManager.addFileManager(getStylesFileManager(less));
  }

};

const loader$4 = {
  name: "less",
  test: /\.less$/i,

  async process({
    code,
    map
  }) {
    var _this$options$plugins, _res$map;

    const less = await loadModule("less");
    if (!less) throw new Error("You need to install `less` package in order to process Less files");
    const res = await less.render(code, { ...this.options,
      plugins: [importer$1].concat((_this$options$plugins = this.options.plugins) !== null && _this$options$plugins !== void 0 ? _this$options$plugins : []),
      filename: this.id,
      sourceMap: {
        outputSourceFiles: true,
        sourceMapBasepath: path.dirname(this.id)
      }
    });
    const deps = res.imports;

    for (const dep of deps) this.deps.add(normalizePath(dep));

    return {
      code: res.css,
      map: (_res$map = res.map) !== null && _res$map !== void 0 ? _res$map : map
    };
  }

};

function matchFile(file, condition) {
  if (typeof condition === "function") return condition(file);
  return Boolean(condition === null || condition === void 0 ? void 0 : condition.test(file));
} // This queue makes sure one thread is always available,
// which is necessary for some cases
// ex.: https://github.com/sass/node-sass/issues/857


const threadPoolSize = process.env.UV_THREADPOOL_SIZE ? Number.parseInt(process.env.UV_THREADPOOL_SIZE) : 4; // default `libuv` threadpool size

const workQueue = new PQueue({
  concurrency: threadPoolSize - 1
});
class Loaders {
  constructor(options) {
    this.loaders = [];
    this.use = options.use.map(rule => {
      if (typeof rule === "string") return [rule, {}];
      if (Array.isArray(rule)) return rule.length === 1 ? [rule[0], {}] : rule;
      throw new TypeError("The rule in `use` option must be string or array!");
    });

    this.test = file => options.extensions.some(ext => file.toLowerCase().endsWith(ext));

    this.listLoader(loader);
    this.listLoader(loader$1);
    this.listLoader(loader$2);
    this.listLoader(loader$3);
    this.listLoader(loader$4);

    for (const loader of options.loaders) this.listLoader(loader);
  }

  getLoader(name) {
    return this.loaders.find(loader => loader.name === name);
  }

  listLoader(loader) {
    if (!this.use.some(rule => rule[0] === loader.name)) return;
    if (this.getLoader(loader.name)) this.unlistLoader(loader.name);
    this.loaders.push(loader);
  }

  unlistLoader(name) {
    this.loaders = this.loaders.filter(loader => loader.name !== name);
  }

  isSupported(file) {
    return this.test(file) || this.loaders.some(loader => matchFile(file, loader.test));
  }

  async process(payload, context) {
    return this.use.slice().reverse().map(([name, options]) => {
      const loader = this.getLoader(name);
      const ctx = { ...context,
        options: options !== null && options !== void 0 ? options : {}
      };
      return async payload => {
        if (loader && (loader.alwaysProcess || matchFile(ctx.id, loader.test))) {
          return workQueue.add(loader.process.bind(ctx, payload));
        }

        return payload;
      };
    }).reduce(async (current, next) => current.then(next), Promise.resolve({ ...payload
    }));
  }

}

function inferOption(option, defaultValue) {
  if (typeof option === "boolean") return option && {};
  if (typeof option === "object") return option;
  return defaultValue;
}
function inferModeOption(mode) {
  var _mode$, _mode$2;

  const m = Array.isArray(mode) ? {
    inject: mode[0] === "inject" && ((_mode$ = mode[1]) !== null && _mode$ !== void 0 ? _mode$ : true),
    extract: mode[0] === "extract" && ((_mode$2 = mode[1]) !== null && _mode$2 !== void 0 ? _mode$2 : true),
    emit: mode[0] === "emit"
  } : {
    inject: mode === "inject",
    extract: mode === "extract",
    emit: mode === "emit"
  };
  if (!(m.inject || m.extract || m.emit)) m.inject = true;
  return m;
}
function inferSourceMapOption(sourceMap) {
  return Array.isArray(sourceMap) ? Boolean(sourceMap[0]) && {
    content: true,
    ...sourceMap[1],
    inline: sourceMap[0] === "inline"
  } : Boolean(sourceMap) && {
    content: true,
    inline: sourceMap === "inline"
  };
}
function inferHandlerOption(option, alias) {
  const opt = inferOption(option, {
    alias
  });
  if (typeof opt === "object" && !opt.alias) opt.alias = alias;
  return opt;
}
function ensureUseOption(use, opts) {
  var _opts$sass, _opts$less, _opts$stylus;

  const all = {
    sass: ["sass", (_opts$sass = opts.sass) !== null && _opts$sass !== void 0 ? _opts$sass : {}],
    less: ["less", (_opts$less = opts.less) !== null && _opts$less !== void 0 ? _opts$less : {}],
    stylus: ["stylus", (_opts$stylus = opts.stylus) !== null && _opts$stylus !== void 0 ? _opts$stylus : {}]
  };
  if (!Array.isArray(use)) return Object.values(all);
  return use.map(loader => {
    var _all$loader;

    return (_all$loader = all[loader]) !== null && _all$loader !== void 0 ? _all$loader : loader;
  });
}
function ensurePCSSOption(option, type) {
  if (typeof option !== "string") return option;

  try {
    return require(option);
  } catch {
    throw new Error(`Unable to load PostCSS ${type} \`${option}\``);
  }
}
function ensurePCSSPlugins(plugins) {
  if (!Array.isArray(plugins)) return ensurePCSSOption(plugins, "plugin");
  return plugins.map(p => {
    if (isNullish(p)) return;
    if (!Array.isArray(p)) return ensurePCSSOption(p, "plugin");
    const [plug, opts] = p;
    if (!opts) return ensurePCSSOption(plug, "plugin");
    return ensurePCSSOption(plug, "plugin")(opts);
  }).filter(nullishFilter);
}

/// <reference types="./shims/postcss-modules" />
var index = ((options = {}) => {
  var _options$namedExports, _options$autoModules, _options$extensions, _options$loaders;

  const isIncluded = createFilter(options.include, options.exclude);
  const sourceMap = inferSourceMapOption(options.sourceMap);
  const postcssOpts = { ...inferModeOption(options.mode),
    minimize: inferOption(options.minimize, false),
    config: inferOption(options.config, {}),
    import: inferHandlerOption(options.import, options.alias),
    url: inferHandlerOption(options.url, options.alias),
    modules: inferOption(options.modules, false),
    to: options.to,
    namedExports: (_options$namedExports = options.namedExports) !== null && _options$namedExports !== void 0 ? _options$namedExports : false,
    autoModules: (_options$autoModules = options.autoModules) !== null && _options$autoModules !== void 0 ? _options$autoModules : false,
    extensions: (_options$extensions = options.extensions) !== null && _options$extensions !== void 0 ? _options$extensions : [".css", ".pcss", ".postcss", ".sss"],
    postcss: {
      parser: ensurePCSSOption(options.parser, "parser"),
      syntax: ensurePCSSOption(options.syntax, "syntax"),
      stringifier: ensurePCSSOption(options.stringifier, "stringifier"),
      plugins: ensurePCSSPlugins(options.plugins)
    }
  };
  const loaders = new Loaders({
    use: [["postcss", postcssOpts], ...ensureUseOption(options.use, options), "sourcemap"],
    loaders: (_options$loaders = options.loaders) !== null && _options$loaders !== void 0 ? _options$loaders : [],
    extensions: postcssOpts.extensions
  });
  const extracted = new Map();
  let preserveModules;
  const plugin = {
    name: "styles",

    buildStart(opts) {
      preserveModules = Boolean(opts.preserveModules);
    },

    async transform(code, id) {
      var _ref;

      if (!isIncluded(id) || !loaders.isSupported(id)) return null; // Check if file was already processed into JS
      // by other instance(s) of this or other plugin(s)

      try {
        this.parse(code, {}); // If it doesn't throw...

        this.warn(`Skipping processed file ${humanlizePath(id)}`);
        return null;
      } catch {// Was not already processed, continuing
      }

      if (typeof options.onImport === "function") options.onImport(code, id);
      const ctx = {
        id,
        sourceMap,
        deps: new Set(),
        assets: new Map(),
        warn: this.warn.bind(this),
        plugin: this,
        options: {}
      };
      const res = await loaders.process({
        code
      }, ctx);

      for (const dep of ctx.deps) this.addWatchFile(dep);

      for (const [fileName, source] of ctx.assets) this.emitFile({
        type: "asset",
        fileName,
        source
      });

      if (res.extracted) extracted.set(id, res.extracted);
      return {
        code: res.code,
        map: (_ref = sourceMap ? res.map : undefined) !== null && _ref !== void 0 ? _ref : {
          mappings: ""
        },
        moduleSideEffects: Boolean(postcssOpts.extract) || null
      };
    },

    augmentChunkHash(chunk) {
      if (extracted.size === 0) return;
      const ids = Object.keys(chunk.modules).reduce((acc, id) => {
        const i = this.getModuleInfo(id);
        return [...acc, i.id, ...i.importedIds];
      }, []);
      const hashable = [...extracted.values()].filter(e => ids.includes(e.id)).sort((a, b) => ids.lastIndexOf(a.id) - ids.lastIndexOf(b.id)).map(e => ({ ...e,
        id: path.basename(e.id),
        map: mm(e.map).relative(path.dirname(e.id)).toString()
      }));
      if (hashable.length === 0) return;
      return JSON.stringify(hashable);
    },

    async generateBundle(opts, bundle) {
      var _opts$dir, _opts$file;

      if (extracted.size === 0 || !(opts.dir || opts.file)) return;
      const dir = (_opts$dir = opts.dir) !== null && _opts$dir !== void 0 ? _opts$dir : path.dirname((_opts$file = opts.file) !== null && _opts$file !== void 0 ? _opts$file : "");

      const getExtractedData = (name, ids) => {
        const fileName = typeof postcssOpts.extract === "string" ? normalizePath(postcssOpts.extract).replace(/^\.[/\\]/, "") : normalizePath(`${name}.css`);
        if (isAbsolutePath(fileName)) this.error(["Extraction path must be relative to the output directory,", `which is ${humanlizePath(dir)}`].join("\n"));
        if (isRelativePath(fileName)) this.error(["Extraction path must be nested inside output directory,", `which is ${humanlizePath(dir)}`].join("\n"));
        const fileDir = path.dirname(path.resolve(dir, fileName));
        const entries = [...extracted.values()].filter(e => ids.includes(e.id)).sort((a, b) => ids.lastIndexOf(a.id) - ids.lastIndexOf(b.id));
        const concat = new Concat(true, path.basename(fileName), "\n");

        for (const res of entries) {
          const relative = relativePath(dir, res.id);
          const map = mm(res.map).relative(fileDir).toObject();
          concat.add(relative, res.css, map);
        }

        return {
          css: concat.content.toString(),
          map: concat.sourceMap,
          name: fileName
        };
      };

      const getName = chunk => {
        if (opts.file) return path.basename(opts.file, path.extname(opts.file));
        if (preserveModules) return path.basename(chunk.fileName, path.extname(chunk.fileName));
        return chunk.name;
      };

      const getImports = chunk => {
        const traversed = new Set();
        const ordered = [];
        let ids = [];

        for (const module of Object.keys(chunk.modules)) {
          traversed.clear();
          ids.push(module);

          while (ids.length > 0) {
            ids = ids.reduce((acc, id) => {
              if (!isIncluded(id) || traversed.has(id)) return acc;
              if (extracted.has(id)) ordered.push(id);else traversed.add(id);
              return [...acc, ...this.getModuleInfo(id).importedIds];
            }, []);
          }
        }

        return ordered;
      };

      const getEmitted = () => {
        const emittedMap = new Map();
        const chunks = Object.values(bundle).filter(c => c.type === "chunk");
        const entries = chunks.filter(c => c.isEntry || c.isDynamicEntry);

        if (typeof postcssOpts.extract === "string") {
          var _entries$find;

          const name = getName((_entries$find = entries.find(e => e.isEntry)) !== null && _entries$find !== void 0 ? _entries$find : entries[0]);
          const emitted = preserveModules ? chunks : entries;
          const ids = emitted.reduce((acc, e) => [...acc, ...getImports(e)], []);
          if (ids.length !== 0) emittedMap.set(name, ids);
          return emittedMap;
        }

        const moved = [];
        const virtuals = chunks.filter(c => !c.facadeModuleId);

        for (const chunk of virtuals) {
          const name = getName(chunk);
          const ids = getImports(chunk);
          if (ids.length !== 0) emittedMap.set(name, ids);
          moved.push(...ids);
        } // `preserveModules` does not support the `manualChunks` option,
        // so there will be no overlap


        const emitted = preserveModules ? chunks : entries;

        for (const chunk of emitted) {
          const name = getName(chunk);
          const ids = getImports(chunk).filter(id => !moved.includes(id));
          if (ids.length !== 0) emittedMap.set(name, ids);
        }

        return emittedMap;
      };

      for await (const [name, ids] of getEmitted()) {
        const res = getExtractedData(name, ids);

        if (typeof options.onExtract === "function") {
          const shouldExtract = options.onExtract(res);
          if (!shouldExtract) continue;
        } // Perform minimization on the extracted file


        if (postcssOpts.minimize) {
          var _resMin$map;

          const cssNanoOpts = typeof postcssOpts.minimize === "object" ? postcssOpts.minimize : {};
          cssNanoOpts.from = res.name;
          cssNanoOpts.to = res.name;
          cssNanoOpts.map = sourceMap && {
            inline: false,
            annotation: false,
            sourcesContent: sourceMap.content,
            prev: res.map
          };
          const resMin = await cssnano.process(res.css, cssNanoOpts);
          res.css = resMin.css;
          res.map = (_resMin$map = resMin.map) === null || _resMin$map === void 0 ? void 0 : _resMin$map.toString();
        }

        const cssFileId = this.emitFile({
          type: "asset",
          name: res.name,
          source: res.css
        });

        if (res.map && sourceMap) {
          const fileName = this.getFileName(cssFileId);
          const assetDir = opts.assetFileNames ? normalizePath(path.dirname(opts.assetFileNames)) : "assets"; // Default for Rollup v2

          const map = mm(res.map).modify(m => {
            m.file = path.basename(fileName);
          }).modifySources(s => {
            // Compensate for possible nesting depending on `assetFileNames` value
            if (s === "<no source>") return s;
            if (assetDir.length <= 1) return s;
            s = `../${s}`; // If it didnt return then there's definitely at least 1 level offset

            for (const c of assetDir) if (c === "/") s = `../${s}`;

            return s;
          });

          if (sourceMap.inline) {
            bundle[fileName].source += map.toCommentData();
          } else {
            const mapFileId = this.emitFile({
              type: "asset",
              fileName: `${fileName}.map`,
              source: map.toString()
            });
            const mapFileName = path.basename(this.getFileName(mapFileId));
            bundle[fileName].source += map.toCommentFile(mapFileName);
          }
        }
      }
    }

  };
  return plugin;
});

export default index;
